/************************************************************************
 *
 *   Sketchup/Python - Mapping of Ruby exceptions to Python exceptions
 *
 ************************************************************************/

#include <stdarg.h>
#include "ruby-python.h"
#include "rubyexc.h"

static const int debug = 0;
#define DEBUGP if (debug) printf

PyObject *PyExc_RubyError;

/*----------------------------------------------------------------------------*/

struct iter_method_arg {
    VALUE obj;
    ID mid;
    int argc;
    VALUE *argv;
};

static VALUE iterate_method _((VALUE obj));
static VALUE
iterate_method(obj)
    VALUE obj;
{
    struct iter_method_arg *arg;

    arg = (struct iter_method_arg *)obj;
    return rb_funcall2(arg->obj, arg->mid, arg->argc, arg->argv);
}

static VALUE
_rb_block_call(obj, mid, argc, argv, bl_proc, data2)
    VALUE obj;
    ID mid;
    int argc;
    VALUE *argv;
    VALUE (*bl_proc) (ANYARGS);
    VALUE data2;
{
    struct iter_method_arg arg;

    arg.obj = obj;
    arg.mid = mid;
    arg.argc = argc;
    arg.argv = argv;
    return rb_iterate(iterate_method, (VALUE)&arg, bl_proc, data2);
}

/*----------------------------------------------------------------------------*/

struct funcallargs {
	VALUE obj;
	ID id;
	int argc;
	VALUE *argv;
	VALUE (*bfunc)();
	VALUE bdata;
};

static VALUE vfuncall2(VALUE ap) {
	struct funcallargs *fa = (struct funcallargs *)ap;
	return rb_funcall2(fa->obj, fa->id, fa->argc, fa->argv);
}

static VALUE vblockcall2(VALUE ap) {
	struct funcallargs *fa = (struct funcallargs *)ap;
	return _rb_block_call(fa->obj, fa->id, fa->argc, fa->argv, fa->bfunc, fa->bdata);
}

VALUE protect_funcall2(VALUE obj, ID id, int argc, VALUE argv[], int *status) {
	struct funcallargs args;
	VALUE result;
	
	args.obj = obj;
	args.id = id;
	args.argc = argc;
	args.argv = argv;
	return rb_protect(vfuncall2, (VALUE)&args, status);
}

VALUE protect_blockcall2(VALUE obj, ID id, int argc, VALUE argv[],
		VALUE (*func)(), VALUE data, int *status)
{
	struct funcallargs args;
	VALUE result;
	
	args.obj = obj;
	args.id = id;
	args.argc = argc;
	args.argv = argv;
	args.bfunc = func;
	args.bdata = data;
	return rb_protect(vblockcall2, (VALUE)&args, status);
}

VALUE protect_funcall(int *status, VALUE obj, ID id, int argc, ...) {
	VALUE ret;
	
	va_list ap;
	va_start(ap, argc);
	ret = protect_funcall2(obj, id, argc, (VALUE *)ap, status);
	va_end(ap);
	return ret;
}

/*----------------------------------------------------------------------------*/

struct func_value_id {
	value_id_func func;
	VALUE value;
	ID id;
};

VALUE apply_value_id(struct func_value_id *args) {
	return args->func(args->value, args->id);
}

VALUE protect_value_id(value_id_func func, VALUE value, ID id, int *status) {
	struct func_value_id args;
	VALUE ret;
	
	args.func = func;
	args.value = value;
	args.id = id;
	return rb_protect((VALUE(*)())apply_value_id, (VALUE)&args, status);
}

/*----------------------------------------------------------------------------*/

static int valid_python_exception(PyObject *obj) {
#if PY_VERSION_HEX >= 0x02050000
	return PyObject_IsInstance(obj, PyExc_BaseException);
#else
	return PyInstance_Check(obj);
#endif
}

void raise_py() {
	VALUE exc;
	PyObject *py_exc = 0;
	PyObject *py_type;
	char *msg;
	
	DEBUGP("--> raise_py\n");
	exc = rb_gv_get("$!");
	rb_gv_set("$!", Qnil);
	if (rb_obj_is_kind_of(exc, cPyObject)) {
		DEBUGP("... is a cPyObject\n");
		Data_Get_Struct(exc, PyObject, py_exc);
		if (valid_python_exception(py_exc)) {
			DEBUGP("... is a Python exception\n");
			Py_INCREF(py_exc);
		}
		else
			py_exc = 0;
	}
	if (!py_exc) {
		DEBUGP("... is a Ruby exception\n");
		msg = RSTRING(rb_inspect(exc))->ptr;
		DEBUGP("... PyObject_CallFunction on %p with %s\n", PyExc_RubyError, msg);
		py_exc = PyObject_CallFunction(PyExc_RubyError, "s", msg);
	}
	if (PyInstance_Check(py_exc))
		py_type = (PyObject *)((PyInstanceObject *)py_exc)->in_class;
	else
		py_type = (PyObject *)py_exc->ob_type;
	PyErr_SetObject(py_type, py_exc);
	Py_DECREF(py_exc);
	DEBUGP("<-- raise_py\n");
}

PyObject *rb2py_funcall(VALUE obj, ID id, int argc, ...) {
	va_list ap;
	PyObject *ret;

	va_start(ap, argc);
	ret = rb2py_funcall2(obj, id, argc, (VALUE *)ap);
	va_end(ap);
	return ret;
}

PyObject *rb2py_funcall2(VALUE obj, ID id, int argc, VALUE argv[]) {
	VALUE ret;
	int status;

	ret = protect_funcall2(obj, id, argc, argv, &status);
	if (status) {
		raise_py();
		return 0;
	}
	return rb2py(ret);
}

PyObject *rb2py_blockcall2(VALUE obj, ID id, int argc, VALUE argv[],
	VALUE (*func)(), VALUE data)
{
	VALUE ret;
	int status;

	ret = protect_blockcall2(obj, id, argc, argv, func, data, &status);
	if (status) {
		raise_py();
		return 0;
	}
	return rb2py(ret);
}

int num2int_funcall(VALUE obj, ID id, int argc, ...) {
	va_list ap;
	VALUE ret;
	int status;
	
	va_start(ap, argc);
	ret = protect_funcall2(obj, id, argc, (VALUE *)ap, &status);
	if (status) {
		raise_py();
		return -1;
	}
	return NUM2INT(ret);
}

int rb2void_funcall(VALUE obj, ID id, int argc, ...) {
	va_list ap;
	int status;
	
	va_start(ap, argc);
	protect_funcall2(obj, id, argc, (VALUE *)ap, &status);
	if (status) {
		raise_py();
		return -1;
	}
	return 0;
}

PyObject *rb2py_value_id(value_id_func func, VALUE value, ID id) {
	VALUE ret;
	int status;
	
	ret = protect_value_id(func, value, id, &status);
	if (status) {
		raise_py();
		return 0;
	}
	return rb2py(ret);
}
