/*
 *	rubymodule.c
 *
 *	$Id: rubymodule.c,v 1.2 1999/11/08 05:45:00 fukusima Exp $
 *
 *	Copyright (C) 1999	Masaki Fukushima
 *
 *	Implements 'ruby' module for Python.
 */

#include "ruby-python.h"
#include "rubyexc.h"
#include "rubymodule.h"

static const int debug = 0;
#define DEBUGP if (debug) printf

VALUE
RbObject_GetValue(self)
	PyObject *self;
{
	if (!RbObject_Check(self))
		rb_bug("not RbObject");
	return ((RbObject *)self)->value;
}

/*
 * Implementation of RbObject.
 */

static PyObject *RbMethod_New(RbObject *, ID);
static RbObject object_list;

PyObject *
RbObject_New(value)
	VALUE value;
{
	RbObject *self;

	self = PyObject_NEW(RbObject, &RbObject_Type);
	if (self == NULL)
		return NULL;

	//printf("--> RbObject_New %p %d\n", self, value);
	//printf("... object_list @%p: %p %p\n", &object_list, object_list.prev, object_list.next);
	//fflush(stdout);
	self->value = value;
	self->prev = object_list.prev;
	self->next = &object_list;
	object_list.prev->next = self;
	object_list.prev = self;
	//printf("<-- RbObject_New\n"); fflush(stdout);
	return (PyObject *)self;
}

static void
RbObject_dealloc(self)
	RbObject *self;
{
	//printf("--> RbObject_dealloc %p %d\n", self, self->value);
	//printf("... object_list @%p: %p %p\n", &object_list, object_list.prev, object_list.next);
	//fflush(stdout);
	self->prev->next = self->next;
	self->next->prev = self->prev;
	PyObject_DEL(self);
	//printf("<-- RbObject_dealloc\n"); fflush(stdout);
}

static int
RbObject_print(self, fp, flags)
	RbObject *self;
	FILE *fp;
	int flags;
{
	VALUE ret;
	int status;
	
	ret = rb_protect(rb_obj_as_string, self->value, &status);
	if (status) {
		raise_py();
		return -1;
	}
	fprintf(fp, "%s", RSTRING(ret)->ptr);
	return 0;
}

static PyObject *
RbObject_repr(self)
	RbObject *self;
{
	VALUE s;
	int status;
	
	s = rb_protect(rb_inspect, self->value, &status);
	if (status) {
		raise_py();
		return 0;
	}
	return PyString_FromStringAndSize(RSTRING(s)->ptr, RSTRING(s)->len);
}

static PyObject *
RbObject_str(self)
	RbObject *self;
{
	VALUE s;
	int status;
	
	s = rb_protect(rb_any_to_s, self->value, &status);
	if (status) {
		raise_py();
		return 0;
	}
	return PyString_FromStringAndSize(RSTRING(s)->ptr, RSTRING(s)->len);
}

static PyObject *
unary_op(char *op, RbObject *px) {
	return rb2py_funcall(px->value, rb_intern(op), 0);
}

static PyObject *
binary_op(char *op, PyObject *px, PyObject *py) {
	VALUE x, y;
	int status;
	
	x = py2rb(px);
	y = py2rb(py);
	return rb2py_funcall(x, rb_intern(op), 1, y);
}

static PyObject *
RbObject_add(PyObject *x, PyObject *y) {
	return binary_op("+", x, y);
}

static PyObject *
RbObject_subtract(PyObject *x, PyObject *y) {
	return binary_op("-", x, y);
}

static PyObject *
RbObject_multiply(PyObject *x, PyObject *y) {
	return binary_op("*", x, y);
}

static PyObject *
RbObject_divide(PyObject *x, PyObject *y) {
	return binary_op("/", x, y);
}

static PyObject *
RbObject_remainder(PyObject *x, PyObject *y) {
	return binary_op("%", x, y);
}

static PyObject *
RbObject_negative(RbObject *x) {
	return unary_op("-@", x);
}

static int
RbObject_compare(self, other)
	RbObject *self;
	PyObject *other;
{
	VALUE v, other_value;
	int id, d;
	int status;
	
	if ((PyObject *)self == other)
		return 0;
	//printf("--> RbObject_compare\n");
	id = rb_intern("<=>");
	if (!rb_respond_to(self->value, id)) {
		if (RbObject_Check(other)) {
			other_value = ((RbObject *)other)->value;
			//printf("... comparing VALUEs %ld %ld\n", self->value, other_value);
			d = self->value - other_value;
			return d > 0 ? 1 : d < 0 ? -1 : 0;
		}
		else
			return -1;
	}
	v = protect_funcall(&status, self->value, id, 1, py2rb(other));
	if (status) {
		raise_py();
		return -1;
	}
	return FIX2INT(v);
}

static int
RbObject_hash(self)
	RbObject *self;
{
	return num2int_funcall(self->value, rb_intern("hash"), 0);
}

static PyObject *
RbObject_insert(self, args)
	RbObject *self;
	PyObject *args;
{
	PyObject *item;
	int index;
	VALUE ret;

	if (!PyArg_ParseTuple(args, "iO", &index, &item))
		return NULL;
	return rb2py_funcall(self->value, rb_intern("[]="),
		3, INT2NUM(index), INT2FIX(0), py2rb(item));
}

static PyMethodDef RbObject_methods[] = {
	{"insert",	(PyCFunction)RbObject_insert,	1},
	{NULL,				NULL}			/* sentinel */
};

static PyObject *
py_bases(VALUE rclass) {
	VALUE rsuper;
	PyObject *psuper = 0, *ret = 0;
	
	//printf("--> py_bases\n"); fflush(stdout);
	rsuper = rb_funcall(rclass, rb_intern("superclass"), 0);
	//printf("... rsuper = %p %s\n", rsuper,
	//	RSTRING(rb_inspect(rsuper))->ptr); fflush(stdout);
	if (rsuper == Qnil) {
		//printf("... rsuper is nil, creating empty ret tuple\n"); fflush(stdout);
		ret = PyTuple_New(0);
		//printf("... ret = %p\n", ret); fflush(stdout);
	}
	else {
		//printf("... converting rsuper to Python object\n"); fflush(stdout);
		psuper = rb2py(rsuper);
		//printf("... psuper = %p\n", psuper); fflush(stdout);
		if (!psuper)
			goto bad;
		//printf("... creating ret tuple\n"); fflush(stdout);
		ret = PyTuple_New(1);
		if (!ret)
			goto bad;
		//printf("... ret = %p\n", ret); fflush(stdout);
		PyTuple_SET_ITEM(ret, 0, psuper);
		//printf("... set item\n"); fflush(stdout);
	}
	//printf("<-- py_bases\n"); fflush(stdout);
	return ret;

bad:
	Py_XDECREF(psuper);
	Py_XDECREF(ret);
	return 0;
}

static PyObject *
py_methods(VALUE rself) {
	VALUE rmeths;	
	ID id;
	PyObject *parray = 0;
	PyObject *plist = 0;
	PyObject *piter = 0;
	PyObject *pitem = 0;

	if (TYPE(rself) == T_CLASS)
		id = rb_intern("instance_methods");
	else if (TYPE(rself) == T_MODULE)
		id = rb_intern("singleton_methods");
	else
		id = rb_intern("methods");
	parray = rb2py_funcall(rself, id, 0);
	plist = PyList_New(0);
	piter = PyObject_GetIter(parray);
	while (pitem = PyIter_Next(piter)) {
		if (PyList_Append(plist, pitem) < 0)
			goto bad;
		Py_DECREF(pitem);
		pitem = 0;
	}
	if (PyErr_Occurred())
		goto bad;
	Py_DECREF(piter);
	Py_DECREF(parray);
	return plist;

bad:
	Py_XDECREF(parray);
	Py_XDECREF(plist);
	Py_XDECREF(piter);
	Py_XDECREF(pitem);
	return 0;
}

static PyObject *
RbObject_getattr(self, name)
	RbObject *self;
	char *name;
{
	PyObject *m;
	int id;
	int len;
	char *name2;
	char buf[BUFSIZ];

	/*printf("RbObject_getattr: '%s' of %s\n", name,
		RSTRING(rb_inspect(self->value))->ptr);*/
	
	if (strcmp(name, "__class__") == 0)
		return rb2py_funcall(self->value, rb_intern("class"), 0);
	
	if (strcmp(name, "__bases__") == 0 && TYPE(self->value) == T_CLASS)
		return py_bases(self->value);
	
	if (strcmp(name, "__methods__") == 0)
		return py_methods(self->value);

	m = Py_FindMethod(RbObject_methods, (PyObject *)self, name);
	if (m != NULL)
		return m;
	PyErr_Clear();
	id = rb_intern(name);

	/* Look for a constant */
	if ((TYPE(self->value) == T_MODULE || TYPE(self->value) == T_CLASS)
		&& rb_is_const_id(id))
			return rb2py_value_id(rb_const_get_at, self->value, id);

	/* Try the name as-is */
	if (rb_respond_to(self->value, id))
		goto found;

	/* Try variations on the name */
	len = strlen(name);
	name2 = ALLOCA_N(char, len + 2);
	
	/* Map is_... to ...? */
	if (strncmp(name, "is_", 3) == 0) {
		strcpy(name2, name + 3);
		name2[len-3] = '?';
		name2[len-2] = 0;
		id = rb_intern(name2);
		if (rb_respond_to(self->value, id))
			goto found;
		else
			goto notfound;
	}
	
	/* Map ..._ip to ...! */
	if (strcmp(name + len - 3, "_ip") == 0) {
		strncpy(name2, name, len - 3);
		name2[len - 3] = '!';
		name2[len - 2] = 0;
		id = rb_intern(name2);
		if (rb_respond_to(self->value, id))
			goto found;
		else
			goto notfound;
	}
	
	/* Add '?' */
	strcpy(name2, name);
	name2[len] = '?';
	name2[len + 1] = '\0';
	id = rb_intern(name2);
	if (rb_respond_to(self->value, id))
		goto found;
	
	/* Add '!' */
	name2[len] = '!';
	id = rb_intern(name2);
	if (rb_respond_to(self->value, id))
		goto found;

notfound:
	snprintf(buf, BUFSIZ, "undefined method `%s' for %s",
			 name, RSTRING(rb_inspect(self->value))->ptr);
	PyErr_SetString(PyExc_AttributeError, buf);
	return NULL;

found:
	return RbMethod_New(self, id);
}

static int
RbObject_setattr(self, name, v)
	RbObject *self;
	char *name;
	PyObject *v;
{
	char *name2;
	int len;
	ID id;
	VALUE val;
	
	len = strlen(name);
	name2 = ALLOCA_N(char, len + 2);
	strcpy(name2, name);
	name2[len] = '=';
	name2[len + 1] = 0;
	id = rb_intern(name2);
	val = py2rb(v);
	return rb2void_funcall(self->value, id, 1, val);
}

static int
RbObject_length(self)
	RbObject *self;
{
	return num2int_funcall(self->value, rb_intern("length"), 0);
}

static PyObject *
RbObject_concat(self, other)
	RbObject *self;
	PyObject *other;
{
	return rb2py_funcall(self->value, rb_intern("+"), 1, py2rb(other));
}

static PyObject *
RbObject_repeat(self, n)
	RbObject *self;
	int n;
{
	VALUE v;

	return rb2py_funcall(self->value, rb_intern("*"), 1, INT2NUM(n));
}

static PyObject *
RbObject_item(self, index)
	RbObject *self;
	int index;
{
	VALUE v;
	int len;

	len = num2int_funcall(self->value, rb_intern("length"), 0);
	if (len < 0)
		return NULL;
	if (index < 0 || index >= len) {
		PyErr_SetString(PyExc_IndexError, "index out of range");
		return NULL;
	}
	return rb2py_funcall(self->value, rb_intern("[]"), 1, INT2NUM(index));
}

static PyObject *
RbObject_slice(self, i1, i2)
	RbObject *self;
	int i1, i2;
{
	VALUE v;

	return rb2py_funcall(self->value, rb_intern("[]"), 2,
		INT2NUM(i1), INT2NUM(i2-i1));
}

static PyObject *
RbObject_ass_item(self, index, val)
	RbObject *self;
	int index;
	PyObject *val;
{
	VALUE v;

	return rb2py_funcall(self->value, rb_intern("[]="),
		2, INT2NUM(index), py2rb(val));
}

static int
RbObject_ass_slice(self, i1, i2, val)
	RbObject *self;
	int i1, i2;
	PyObject *val;
{
	return rb2void_funcall(self->value, rb_intern("[]="),
		3, INT2NUM(i1), INT2NUM(i2-i1), py2rb(val));
}

static PyObject *
RbObject_subscript(self, key)
	RbObject *self;
	PyObject *key;
{
	return rb2py_funcall(self->value, rb_intern("[]"), 1, py2rb(key));
}

static int
RbObject_ass_sub(self, key, val)
	RbObject *self;
	PyObject *key, *val;
{
	VALUE rval;
	PyObject *ret;

	if (val == NULL)
		rval = Qnil;
	else
		rval = py2rb(val);

	return rb2void_funcall(self->value, rb_intern("[]="), 2,
		py2rb(key), rval);
}

static void
pytuple2rbarg(tuple, argc, argv)
	PyObject *tuple;
	int argc;
	VALUE *argv;
{
	int i;

	for (i = 0; i < argc; i++)
		argv[i] = py2rb(PyTuple_GET_ITEM(tuple, i));
}

static PyObject *
RbObject_call(self, arg, kw)
	RbObject *self;
	PyObject *arg, *kw;
{
	VALUE *argv, ret;
	int argc;
	int id;

	argc = PyTuple_Size(arg);
	argv = ALLOCA_N(VALUE, argc);
	pytuple2rbarg(arg, argc, argv);
	id = rb_intern("call");
	if (!rb_respond_to(self->value, id))
		id = rb_intern("new");
	return rb2py_funcall2(self->value, id, argc, argv);
}

static int
RbObject_nonzero(RbObject *self) {
	int id;
	int val;
	
	if (self->value == Qnil)
		return 0;
	id = rb_intern("size");
	if (!rb_respond_to(self->value, id))
		id = rb_intern("length");
	if (rb_respond_to(self->value, id)) {
		val = num2int_funcall(self->value, id, 0);
		if (val < 0)
			return val;
		return val != 0;
	}
	return 1;
}

static PySequenceMethods rbobject_as_sequence = {
	(lenfunc)RbObject_length,			/*sq_length*/
	(binaryfunc)RbObject_concat,		/*sq_concat*/
	(ssizeargfunc)RbObject_repeat,		/*sq_repeat*/
	(ssizeargfunc)RbObject_item,			/*sq_item*/
	(ssizessizeargfunc)RbObject_slice,		/*sq_slice*/
	(ssizeobjargproc)RbObject_ass_item,	/*sq_ass_item*/
	(ssizessizeobjargproc)RbObject_ass_slice, /*sq_ass_slice*/
};

static PyMappingMethods rbobject_as_mapping = {
	(lenfunc)RbObject_length,			/*mp_length*/
	(binaryfunc)RbObject_subscript,		/*mp_subscript*/
	(objobjargproc)RbObject_ass_sub,	/*mp_ass_subscript*/
};

static PyNumberMethods rbobject_as_number = {
	(binaryfunc)RbObject_add, /* binaryfunc nb_add */
	(binaryfunc)RbObject_subtract, /* binaryfunc nb_subtract */
	(binaryfunc)RbObject_multiply, /* binaryfunc nb_multiply */
	(binaryfunc)RbObject_divide, /* binaryfunc nb_divide */
	(binaryfunc)RbObject_remainder, /* binaryfunc nb_remainder */
	0, /* binaryfunc nb_divmod */
	0, /* ternaryfunc nb_power */
	(unaryfunc)RbObject_negative, /* unaryfunc nb_negative */
	0, /* unaryfunc nb_positive */
	0, /* unaryfunc nb_absolute */
	(inquiry)RbObject_nonzero, /* inquiry nb_nonzero */
	0, /* unaryfunc nb_invert */
	0, /* binaryfunc nb_lshift */
	0, /* binaryfunc nb_rshift */
	0, /* binaryfunc nb_and */
	0, /* binaryfunc nb_xor */
	0, /* binaryfunc nb_or */
	0, /* coercion nb_coerce */
	0, /* unaryfunc nb_int */
	0, /* unaryfunc nb_long */
	0, /* unaryfunc nb_float */
	0, /* unaryfunc nb_oct */
	0, /* unaryfunc nb_hex */

	/* Added in release 2.0 */
	0, /* binaryfunc nb_inplace_add */
	0, /* binaryfunc nb_inplace_subtract */
	0, /* binaryfunc nb_inplace_multiply */
	0, /* binaryfunc nb_inplace_divide */
	0, /* binaryfunc nb_inplace_remainder */
	0, /* ternaryfunc nb_inplace_power */
	0, /* binaryfunc nb_inplace_lshift */
	0, /* binaryfunc nb_inplace_rshift */
	0, /* binaryfunc nb_inplace_and */
	0, /* binaryfunc nb_inplace_xor */
	0, /* binaryfunc nb_inplace_or */

	/* Added in release 2.2 */
	/* The following require the Py_TPFLAGS_HAVE_CLASS flag */
	(binaryfunc)RbObject_divide, /* binaryfunc nb_floor_divide */
	(binaryfunc)RbObject_divide, /* binaryfunc nb_true_divide */
	0, /* binaryfunc nb_inplace_floor_divide */
	0, /* binaryfunc nb_inplace_true_divide */
};

PyTypeObject RbObject_Type = {
	PyObject_HEAD_INIT(0)             /* (&PyType_Type) */
	0,					                      /*ob_size*/
	"ruby object",		                /*tp_name*/
	sizeof(RbObject),	                /*tp_basicsize*/
	0,					                      /*tp_itemsize*/
	/* methods */
	(destructor)RbObject_dealloc,		  /*tp_dealloc*/
	(printfunc)RbObject_print,			  /*tp_print*/
	(getattrfunc)RbObject_getattr,		/*tp_getattr*/
	(setattrfunc)RbObject_setattr,		/*tp_setattr*/
	(cmpfunc)RbObject_compare,			  /*tp_compare*/
	(reprfunc)RbObject_repr,          /*tp_repr*/
	&rbobject_as_number,							/*tp_as_number*/
	&rbobject_as_sequence,				    /*tp_as_sequence*/
	&rbobject_as_mapping,				      /*tp_as_mapping*/
	(hashfunc)RbObject_hash,					/*tp_hash*/
	(ternaryfunc)RbObject_call,			  /*tp_call*/
	(reprfunc)RbObject_str,           /*tp_str*/
};

/*
 * Ruby method object.
 */
typedef struct RbMethod {
	PyObject_HEAD
	RbObject *self;
	ID id;
} RbMethod;

static RbMethod *meth_free_list = NULL;
static PyTypeObject RbMethod_Type;

static PyObject *
RbMethod_New(self, id)
	RbObject *self;
	ID id;
{
	RbMethod *meth;

	if (meth_free_list == NULL) {
		meth = PyObject_NEW(RbMethod, &RbMethod_Type);
		if (meth == NULL)
			return NULL;
	} else {
		meth = meth_free_list;
		meth_free_list = (RbMethod *)meth->self;
		meth->ob_type = &RbMethod_Type;
		_Py_NewReference((PyObject *)meth);
	}

	Py_INCREF(self);
	meth->self = self;
	meth->id = id;
	return (PyObject *)meth;
}

static void
RbMethod_dealloc(meth)
	RbMethod *meth;
{
	Py_DECREF(meth->self);
	meth->self = (RbObject *)meth_free_list;
	meth_free_list = meth;
}

static VALUE
call_py_block(VALUE arg, VALUE r_block) {
	PyObject *py_block = 0;
	PyObject *py_arg = 0;
	PyObject *py_args = 0;
	PyObject *py_ret = 0;
	VALUE ret;

	py_block = rb2py(r_block);
	if (arg == Qnil)
		py_args = PyTuple_New(0);
	else if (TYPE(arg) == T_ARRAY) {
		int len = RARRAY(arg)->len;
		int i;
		py_args = PyTuple_New(len);
		if (!py_args)
			goto bad;
		for (i = 0; i < len; i++) {
			py_arg = rb2py_nr(RARRAY(arg)->ptr[i]);
			if (!py_arg)
				goto bad;
			PyTuple_SET_ITEM(py_args, i, py_arg);
			py_arg = 0;
		}
	}
	else {
		py_arg = rb2py(arg);
		py_args = PyTuple_New(1);
		if (!py_args)
			goto bad;
		PyTuple_SET_ITEM(py_args, 0, py_arg);
		py_arg = 0;
	}
	py_ret = PyObject_Call(py_block, py_args, 0);
	if (!py_ret)
		goto bad;
	Py_DECREF(py_block);
	Py_DECREF(py_args);
	ret = py2rb_decref(py_ret);
	return ret;

bad:
	Py_XDECREF(py_block);
	Py_XDECREF(py_arg);
	Py_XDECREF(py_args);
	py_raise();
}

static PyObject *
RbMethod_Call(meth, arg, kw)
	RbMethod *meth;
	PyObject *arg, *kw;
{
	VALUE *argv;
	int argc;
	int nkw = 0;
	PyObject *py_block = 0; /* borrowed */
	VALUE r_block;
	PyObject *ret;

	//printf("--> RbMethod_Call\n");
	argc = PyTuple_Size(arg);
	argv = ALLOCA_N(VALUE, argc);
	pytuple2rbarg(arg, argc, argv);
	if (kw) {
		py_block = PyDict_GetItemString(kw, "body");
		if (py_block)
			nkw = 1;
		if (PyObject_Length(kw) != nkw) {
			PyErr_SetString(PyExc_TypeError, "Function does not take keyword arguments");
			goto bad;
		}
	}
	if (py_block) {
		//printf("... body = "); PyObject_Print(py_block, stdout, 0); printf("\n");
		//printf("... calling rb2py_blockcall2\n");
		r_block = py2rb(py_block);
		ret = rb2py_blockcall2(meth->self->value, meth->id, argc, argv,
			call_py_block, r_block);
	}
	else {
		//if (kw) printf("... calling rb2py_funcall2\n");
		ret = rb2py_funcall2(meth->self->value, meth->id, argc, argv);
	}
	//printf("<-- RbMethod_Call\n");
	return ret;

bad:
	return 0;
}

static PyTypeObject RbMethod_Type = {
	PyObject_HEAD_INIT(0) /* (&PyType_Type) */
	0,					/*ob_size*/
	"ruby method",		/*tp_name*/
	sizeof(RbMethod),	/*tp_basicsize*/
	0,					/*tp_itemsize*/
	/* methods */
	(destructor)RbMethod_dealloc,		/*tp_dealloc*/
	0,									/*tp_print*/
	0,									/*tp_getattr*/
	0,									/*tp_setattr*/
	0,									/*tp_compare*/
	0,									/*tp_repr*/
	0,									/*tp_as_number*/
	0,									/*tp_as_sequence*/
	0,									/*tp_as_mapping*/
	0,									/*tp_hash*/
	(ternaryfunc)RbMethod_Call,			/*tp_call*/
};


/*
 * Module "ruby"
 */

/* List of functions defined in ruby module */
static PyMethodDef Rb_methods[] = {
	{NULL,				NULL}			/* sentinel */
};


/*
 * For Ruby GC, all RbObject and RbMethod is maintained in a linked-list.
 */

static void
mark_all_rbobjects(dummy)
	PyObject *dummy;
{
	RbObject *rbobj;
#if DEBUG
	RbMethod *rbmeth;
	int c = 0;
#endif

	/* mark objects referenced from Python */
	rbobj = object_list.next;
	while (rbobj != &object_list) {
		rb_gc_mark(rbobj->value);
		rbobj = rbobj->next;
#if DEBUG
		c++;
#endif
	}

#if DEBUG
	fprintf(stderr, "[%d,", c);
	c = 0;
	rbmeth = (RbMethod *)meth_free_list;
	while (rbmeth != NULL) {
		c++;
		rbmeth = (RbMethod *)rbmeth->self;
	}
	fprintf(stderr, "%d]", c);
	fflush(stderr);
#endif
}

static VALUE dummy_value;

void
initruby()
{
	PyObject *m, *d;
		
	/* Windows can't statically initialize pointers in dlls */
	RbObject_Type.ob_type = &PyType_Type;
	RbMethod_Type.ob_type = &PyType_Type;
	
	PyType_Ready(&RbObject_Type);
	PyType_Ready(&RbMethod_Type);

	PyExc_RubyError = PyErr_NewException("__builtin__.RubyError", 0, 0);

	/* setup lists */
	object_list.value = Qnil;
	object_list.next  = &object_list;
	object_list.prev  = &object_list;

	/* dummy ruby global variable */
	dummy_value = Data_Wrap_Struct(rb_cObject, mark_all_rbobjects, 0, NULL);
	rb_global_variable(&dummy_value);
	
	/* Create the module and add the functions */
	m = Py_InitModule("ruby", Rb_methods);
	PyDict_SetItemString(pdMainDict, "ruby", m);
	d = PyModule_GetDict(m);
	PyDict_SetItemString(d, "Object", (PyObject *)&RbObject_Type);
	PyDict_SetItemString(d, "Method", (PyObject *)&RbMethod_Type);
	PyDict_SetItemString(d, "RubyError", PyExc_RubyError);

}
